﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Linq;
using System.Text.RegularExpressions;
using WMM.Tapioca.Linkbase;
using WMM.Tapioca.Instance;
using WMM.Tapioca.Taxonomy;

namespace WMM.Tapioca.Formula.Filters
{
    /// <summary>
    /// The ancestor filter can be used to select facts that have a specified ancestor element. 
    /// </summary>
    public class AncestorFilter : Filter
    {
        private string ancestor;

        /// <returns> The ancestor element. </returns>
        public string Ancestor
        {
            get { return ancestor; }
            set { ancestor = value; }
        }

        /// <summary>
        /// Constructor.
        /// </summary>
        public AncestorFilter(Resource resource)
            : base(resource)
        {
            XElement ancestorX = resource.ResourceXElement.Descendants().Where(d => d.Parent.Name.LocalName.Equals("ancestor") && 
            d.Name.LocalName.Equals("qname")).FirstOrDefault();

            if (ancestorX != null)
            {
                string qname = ancestorX.Value;
                qname = qname.TrimEnd().TrimStart();
                qname = Regex.Replace(qname, @"\s+", " ");
                ancestor = qname;
            }

            XElement qnameExpressionX = resource.ResourceXElement.Descendants().Where(d => d.Parent.Name.LocalName.Equals("ancestor") &&
            d.Name.LocalName.Equals("qnameExpression")).FirstOrDefault();

            if (qnameExpressionX != null)
            {
                string expression = qnameExpressionX.Value;
                expression = expression.TrimEnd().TrimStart();
                expression = Regex.Replace(expression, @"\s+", " ");
                ancestor = EvaluateExpression(expression);
            }
        }


        public override List<FactItem> EvaluateFilter(VariableSet variableSet, List<FactItem> itemsToFilter, DTS dts)
        {
            this.dts = dts;

            List<FactItem> itemListTemp = new List<FactItem>();

            foreach (FactItem item in itemsToFilter)
            {
                string ancestor = Ancestor.Substring(Ancestor.IndexOf(':') + 1);

                if (Complement ? !(item.FactElement.Ancestors().Where(a => a.Name.LocalName.Equals(ancestor)).Count() > 0) : item.FactElement.Ancestors().Where(a => a.Name.LocalName.Equals(ancestor)).Count() > 0)
                {
                    itemListTemp.Add(item);
                }
            }

            return itemListTemp;
        }
    }
}
