﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Linq;
using System.Text.RegularExpressions;
using WMM.Tapioca.Instance;
using WMM.Tapioca.Taxonomy;

namespace WMM.Tapioca.Formula.Filters
{
    /// <summary>
    /// The concept data-type filter can be used to match facts based upon its XML Schema data type
    /// </summary>
    public class ConceptDataTypeFilter : Filter
    {
        private bool strict;

        /// <returns> When strict="false" allow to compare with data type derived. </returns>
        public bool Strict
        {
            get { return strict; }
            set { strict = value; }
        }

        private string qName;

        /// <returns> The type qname. </returns>
        public string QName
        {
            get { return qName; }
            set { qName = value; }
        }

        /// <summary>
        /// Constructor.
        /// </summary>
        public ConceptDataTypeFilter(Linkbase.Resource resource)
            : base(resource)
        {
            strict = Boolean.Parse((string)resource.ResourceXElement.Attribute("strict") ?? string.Empty);

            XElement qnameExpression = resource.ResourceXElement.Descendants().Where(d => d.Parent.Name.LocalName.Equals("type") &&
            d.Name.LocalName.Equals("qnameExpression")).FirstOrDefault();

            if (qnameExpression != null)
            {
                string expression = qnameExpression.Value;
                expression = expression.TrimEnd().TrimStart();
                expression = Regex.Replace(expression, @"\s+", " ");
                qName = EvaluateExpression(expression);
            }

            XElement qnameX = resource.ResourceXElement.Descendants().Where(d => d.Parent.Name.LocalName.Equals("type") &&
            d.Name.LocalName.Equals("qname")).FirstOrDefault();

            if (qnameX != null)
            {
                string qname = qnameX.Value;
                qname = qname.TrimEnd().TrimStart();
                qname = Regex.Replace(qname, @"\s+", " ");
                qName = qname;
            }
        }


        public override List<FactItem> EvaluateFilter(VariableSet variableSet, List<FactItem> itemsToFilter, DTS dts)
        {
            this.dts = dts;

            List<FactItem> itemListTemp = new List<FactItem>();

            foreach (FactItem item in itemsToFilter)
            {
                string type = QName.Substring(QName.IndexOf(':') + 1);

                if (Strict)
                {
                    if (Complement ? !item.Concept.TypeName.Name.Equals(type) : item.Concept.TypeName.Name.Equals(type))
                    {
                        itemListTemp.Add(item);
                    }
                }
                else
                {
                    if (Complement ? !(item.Concept.SchemaElement.ElementSchemaType.BaseXmlSchemaType.QualifiedName.Name.Equals(type) || item.Concept.TypeName.Name.Equals(type)) :
                        item.Concept.SchemaElement.ElementSchemaType.BaseXmlSchemaType.QualifiedName.Name.Equals(type) || item.Concept.TypeName.Name.Equals(type))
                    {
                        itemListTemp.Add(item);
                    }
                }
            }

            return itemListTemp;
        }
    }
}
