﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Saxon.Api;
using System.Text.RegularExpressions;
using WMM.Tapioca.Instance;
using WMM.Tapioca.Taxonomy;
using WMM.Tapioca.Utilities;


namespace WMM.Tapioca.Formula.Filters
{
    /// <summary>
    /// The general unit filter can be used to select facts based on criteria
    /// that involve a number of unit measures. 
    /// </summary>
    public class GeneralMeasuresFilter : Filter
    {
        private string test;

        /// <returns> The XPath expression implied by a general filter. </returns>
        public string Test
        {
            get { return test; }
            set { test = value; }
        }

        /// <summary>
        /// Constructor.
        /// </summary>
        public GeneralMeasuresFilter(Linkbase.Resource resource)
            : base(resource)
        {
            string expression = (string)resource.ResourceXElement.Attribute("test") ?? string.Empty;
            expression = expression.TrimEnd().TrimStart();
            expression = Regex.Replace(expression, @"\s+", " ");
            test = expression;

        }


        public override List<FactItem> EvaluateFilter(VariableSet variableSet, List<FactItem> itemsToFilter, DTS dts)
        {
            this.dts = dts;

            List<FactItem> itemListTemp = new List<FactItem>();

            foreach (FactItem item in itemsToFilter)
            {
                if (!item.Concept.SchemaElement.ElementSchemaType.IsNumeric())
                {
                    continue;
                }

                string result = EvaluateXPathExpression(variableSet, Test, item.UnitRef.UnitElement.ToXmlNode());

                if (Complement ? !result.Equals("true") : result.Equals("true"))
                {
                    itemListTemp.Add(item);
                }
            }

            return itemListTemp;
        }
    }
}
