﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using WMM.Tapioca.Instance;
using WMM.Tapioca.Taxonomy;
using WMM.Tapioca.Linkbase;
using System.Xml.Linq;

namespace WMM.Tapioca.Formula.Filters
{
    /// <summary>
    /// The complete-scenario matching filter can be used to select facts that have the same scenario,
    /// where the content of the scenario is not interpreted based on the XBRL Dimensions Specification.
    /// </summary>
    public class MatchScenarioFilter : MatchFilter
    {
        /// <summary>
        /// Constructor.
        /// </summary>
        public MatchScenarioFilter(Resource resource)
            : base(resource)
        {
            
        }


        public override List<FactItem> EvaluateFilter(VariableSet variableSet, List<FactItem> itemsToFilter, DTS dts)
        {
            this.dts = dts;
            namespaces = dts.DtsFormula.Namespaces;

            List<FactItem> itemListTemp = new List<FactItem>();

            foreach (FactItem itemToFilter in itemsToFilter)
            {
                List<FactItem> factItems = variableSet.variableItems[Variable];

                for (int i = 0; i < factItems.Count; i++)
                {
                    if (Complement ? !XNode.DeepEquals(itemToFilter.ContextRef.ScenarioElement, factItems[i].ContextRef.ScenarioElement) :
                        XNode.DeepEquals(itemToFilter.ContextRef.ScenarioElement, factItems[i].ContextRef.ScenarioElement))
                    {
                        itemListTemp.Add(itemToFilter);
                        continue;
                    }
                }
            }

            return itemListTemp;
        }
    }
}
