﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Saxon.Api;
using System.Text.RegularExpressions;
using WMM.Tapioca.Instance;
using WMM.Tapioca.Taxonomy;

namespace WMM.Tapioca.Formula.Filters
{
    /// <summary>
    /// The period filter can be used to match facts based upon a broad range of criteria relating
    /// to the period over which or at which they have been measured.
    /// </summary>
    public class Period : Filter
    {
        private string test;

        /// <returns> The XPath expression implied by a general filter. </returns>
        public string Test
        {
            get { return test; }
            set { test = value; }
        }

        /// <summary>
        /// Constructor.
        /// </summary>
        public Period(Linkbase.Resource resource)
            : base(resource)
        {
            string expression = (string)resource.ResourceXElement.Attribute("test") ?? string.Empty;
            expression = expression.TrimEnd().TrimStart();
            expression = Regex.Replace(expression, @"\s+", " ");
            test = expression;
        }


        public override List<FactItem> EvaluateFilter(VariableSet variableSet, List<FactItem> itemsToFilter, DTS dts)
        {
            this.dts = dts;
            namespaces = dts.DtsFormula.Namespaces;

            List<FactItem> itemListTemp = new List<FactItem>();

            foreach (FactItem item in itemsToFilter)
            {
                string result = EvaluateXPathExpression(variableSet, Test, item);

                if (Complement ? !result.Equals("true") : result.Equals("true"))
                {
                    itemListTemp.Add(item);
                }
            }

            return itemListTemp;
        }

    }
}
