﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Linq;
using Saxon.Api;
using System.Text.RegularExpressions;
using WMM.Tapioca.Linkbase;
using WMM.Tapioca.Instance;
using WMM.Tapioca.Taxonomy;

namespace WMM.Tapioca.Formula.Filters
{
    /// <summary>
    /// The period-instant filter can be used to match facts based upon the instant at which they have been measured.
    /// </summary>
    public class PeriodInstantFilter : PeriodFilter
    {
        /// <summary>
        /// Constructor.
        /// </summary>
        public PeriodInstantFilter(Resource resource)
            : base(resource)
        {
            string expression = (string)resource.ResourceXElement.Attribute("date") ?? string.Empty;
            expression = expression.TrimEnd().TrimStart();
            expression = Regex.Replace(expression, @"\s+", " ");
            Date = EvaluateExpression(expression);

            string expTime = (string)resource.ResourceXElement.Attribute("time") ?? string.Empty;
            expTime = expTime.TrimEnd().TrimStart();
            expTime = Regex.Replace(expTime, @"\s+", " ");

            if (expTime != string.Empty)
            {
                Time = EvaluateExpression(expTime);
                DateT = DateTime.Parse(Date + " " + Time);
            }
            else
            {
                Time = "00:00:00";
                DateT = DateTime.Parse(Date + " " + Time);
            }         
        }


        public override List<FactItem> EvaluateFilter(VariableSet variableSet, List<FactItem> itemsToFilter, DTS dts)
        {
            List<FactItem> itemListTemp = new List<FactItem>();            

            foreach (FactItem item in itemsToFilter)
            {
                if (Complement ? !item.ContextRef.InstantDate.Equals(DateT) : item.ContextRef.InstantDate.Equals(DateT))
                {
                    itemListTemp.Add(item);
                }
            }

            return itemListTemp;
        }
    }
}
