﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Linq;
using Saxon.Api;
using System.Text.RegularExpressions;
using WMM.Tapioca.Linkbase;
using WMM.Tapioca.Instance;
using WMM.Tapioca.Taxonomy;
using WMM.Tapioca.Utilities;

namespace WMM.Tapioca.Formula.Filters
{
    /// <summary>
    /// The scenario filter can be used to match facts that have non-XDT content satisfying specified constraints. 
    /// </summary> 
    public class ScenarioFilter : Filter
    {
        private string test;

        /// <returns> The XPath expression implied by a general filter. </returns>
        public string Test
        {
            get { return test; }
            set { test = value; }
        }

        /// <summary>
        /// Constructor.
        /// </summary>
        public ScenarioFilter(Resource resource)
            : base(resource)
        {
            string expression = (string)resource.ResourceXElement.Attribute("test") ?? string.Empty;
            expression = expression.TrimEnd().TrimStart();
            expression = Regex.Replace(expression, @"\s+", " ");
            test = expression;
        }


        public override List<FactItem> EvaluateFilter(VariableSet variableSet, List<FactItem> itemsToFilter, DTS dts)
        {
            this.dts = dts;

            List<FactItem> itemListTemp = new List<FactItem>();

            foreach (FactItem item in itemsToFilter)
            {
                if (item.ContextRef.ScenarioElement == null)
                {
                    continue;
                }

                string result = EvaluateXPathExpression(variableSet, Test, item.ContextRef.ScenarioElement.ToXmlNode());

                if (!result.Equals("Saxon.Api.XdmEmptySequence"))
                {
                    if (Complement ? result.Equals("false") : !result.Equals("false"))
                    {
                        itemListTemp.Add(item);
                    }
                }         
                //else
                //{
                //    if (item.ContextRef.ScenarioElement != null)
                //    {
                //        XElement childElement = item.ContextRef.ScenarioElement.Elements().Where(d => d.Name.LocalName.Equals(result)).FirstOrDefault();

                //        if (childElement != null)
                //        {
                //            itemListTemp.Add(item);
                //        }
                //    }
                //}
            }

            return itemListTemp;
        }
    }
}
