﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using WMM.Tapioca.Instance;
using WMM.Tapioca.Taxonomy;
using WMM.Tapioca.Utilities;

namespace WMM.Tapioca.Formula.Filters
{
    /// <summary>
    /// The specific entity-identifier filter can be used to match facts based upon whether they 
    /// report values using the entity identifier value given by the filter. 
    /// </summary>
    public class SpecificIdentifierFilter : Filter
    {
        private string scheme;

        /// <returns>The scheme is the specific entity identification scheme of the filter. </returns>
        public string Scheme
        {
            get { return scheme; }
            set { scheme = value; }
        }

        private string value;

        /// <returns> The value is the specific entity identification value of the filter. </returns>
        public string Value
        {
            get { return this.value; }
            set { this.value = value; }
        }

        /// <summary>
        /// Constructor.
        /// </summary>
        public SpecificIdentifierFilter(Linkbase.Resource resource)
            : base(resource)
        {
            string expression = (string)resource.ResourceXElement.Attribute("scheme") ?? string.Empty;
            expression = expression.TrimEnd().TrimStart();
            expression = Regex.Replace(expression, @"\s+", " ");
            scheme = expression;

            string valueExpression = (string)resource.ResourceXElement.Attribute("value") ?? string.Empty;
            valueExpression = valueExpression.TrimEnd().TrimStart();
            valueExpression = Regex.Replace(valueExpression, @"\s+", " ");
            this.value = valueExpression;
        }


        public override List<FactItem> EvaluateFilter(VariableSet variableSet, List<FactItem> itemsToFilter, DTS dts)
        {
            this.dts = dts;

            List<FactItem> itemListTemp = new List<FactItem>();

            foreach (FactItem item in itemsToFilter)
            {
                string scheme = EvaluateXPathExpression(Scheme, item.ContextRef.ContextElement.ToXmlNode());
                string value = EvaluateXPathExpression(Value, item.ContextRef.ContextElement.ToXmlNode());

                if (Complement ? !(scheme.Equals(item.ContextRef.IdentifierScheme) && value.Equals(item.ContextRef.Identifier)) :
                    scheme.Equals(item.ContextRef.IdentifierScheme) && value.Equals(item.ContextRef.Identifier))
                {
                    itemListTemp.Add(item);
                }
            }

            return itemListTemp;
        }
    }
}
