﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace WMM.Tapioca.Taxonomy
{
    /// <summary>
    /// A DictionaryMap has multiple values and two keys. 
    /// With it we add multiple values to a single key(Tuple) in a Dictionary
    /// </summary>
    public class DictionaryMap<V>
    {
        Dictionary<string, List<V>> _dictionary =
        new Dictionary<string, List<V>>();

        /// <summary>
        /// This method is implemented with TryGetValue, and therefore
        /// usually has better performance than ContainsKey.
        /// </summary>
        public void Add(string key, V value)
        {
            List<V> list;

            if (this._dictionary.TryGetValue(key, out list))
            {
                list.Add(value);
            }
            else
            {
                list = new List<V>();
                list.Add(value);
                this._dictionary[key] = list;
            }
        }

        /// <summary>
        /// You will want to enumerate over Keys in your code.
        /// </summary>
        public IEnumerable<string> Keys
        {
            get
            {
                return this._dictionary.Keys;
            }
        }

        /// <summary>
        /// This provides natural and short syntax for accessing the DictionaryMap.
        /// Note that it returns an empty List if nothing exists.
        /// </summary>
        public List<V> this[string key]
        {
            get
            {
                List<V> list;

                if (!this._dictionary.TryGetValue(key, out list))
                {
                    list = new List<V>();
                    this._dictionary[key] = list;
                }

                return list;
            }
            set
            {
                List<V> list;

                if (!this._dictionary.TryGetValue(key, out list))
                {
                    list = new List<V>();
                    this._dictionary[key] = list;
                }
            }
        }
    }
}
