using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;

namespace WMM.Tapioca
{
    internal static class XmlUtilities
    {
        //------------------------------------------------------------------------------------
        // Finds an attribute with the given name in the given XML node and returns the
        // attribute of the value. Returns an empty string if no such attribute exists. Works
        // for local names as well as names qualified with a namespace identifier.
        //------------------------------------------------------------------------------------
        internal static string GetAttributeValue(XmlNode node, string attributeName)
        {
            bool nameIncludesNamespaceId;

            if (attributeName.IndexOf(':') == -1)
                nameIncludesNamespaceId = false;
            else
                nameIncludesNamespaceId = true;
            if (node == null)
                return string.Empty;
            if (node.Attributes == null)
                return string.Empty;
            foreach (XmlAttribute currentAttribute in node.Attributes)
            {
                if (nameIncludesNamespaceId == false)
                {
                    if (currentAttribute.LocalName.Equals(attributeName))
                        return currentAttribute.Value;
                }
                else
                {
                    if (currentAttribute.Name.Equals(attributeName))
                        return currentAttribute.Value;
                }
            }
            return string.Empty;
        }

        //------------------------------------------------------------------------------------
        // Finds an attribute with the given namespace URI and the given local name in the
        // given XML node and returns the attribute of the value. Returns an empty string if
        // no such attribute exists.
        //------------------------------------------------------------------------------------
        internal static string GetAttributeValue(XmlNode node, string attributeNamespaceUri, string attributeLocalName)
        {
            if (node == null)
                return string.Empty;
            if (node.Attributes == null)
                return string.Empty; 
            foreach (XmlAttribute currentAttribute in node.Attributes)
            {
                if (currentAttribute.NamespaceURI.Equals(attributeNamespaceUri))
                {
                    if (currentAttribute.LocalName.Equals(attributeLocalName))
                        return currentAttribute.Value;
                }
            }
            return string.Empty;
        }
    }
}
